package preferences;

###############################################################################
#                                                                             #
# Slrnconf - a graphical configuration utility for SLRN                       #
# Copyright (C) 2000-2003 Torsten Schoenfeld                                  #
#                                                                             #
# This program is free software; you can redistribute it and/or modify it     #
# under the terms of the GNU General Public License as published by the Free  #
# Software Foundation; either version 2 of the License, or (at your option)   #
# any later version.                                                          #
#                                                                             #
# This program is distributed in the hope that it will be useful, but WITHOUT #
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       #
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   #
# more details.                                                               #
#                                                                             #
# You should have received a copy of the GNU General Public License along     #
# with this program; if not, write to the Free Software Foundation, Inc., 59  #
# Temple Place, Suite 330, Boston, MA 02111-1307 USA                          #
#                                                                             #
###############################################################################

use strict;
use rc_grammar;

sub new {
  my ($class, $ui) = @_;
  my $self = {};

  bless($self, $class);

  $self -> { rc_file } = $ENV{ HOME } . "/.slrnrc";
  $self -> { ui } = $ui;

  $self -> load_defaults();

  return $self;
}

sub new_key {
  return bless({}, shift());
}

sub add {
  my $self = shift();
  my $key = shift();

  unless (exists($self -> { _orignal_caller })) {
    $self -> { _original_caller } = scalar(caller());
  }

  unless ($#_ == 0) {
    unless (exists($self -> { $key })) {
      $self -> { $key } = preferences -> new_key();
    }

    $self -> { $key } -> add(@_);
  }
  else {
    push(@{$self -> { $key }}, shift());

    unless ($self -> { _original_caller } eq "ui") {
      if (exists($self -> { $key . "_callback" })) {
        $self -> { $key . "_callback" } -> ($self -> { $key });
      }
    }

    delete($self -> { _original_caller });
  }

  $self -> { _changed } = 1 if (caller() eq "ui");

  return 1;
}

sub append {
  my $self = shift();
  my $key = shift();

  unless (exists($self -> { _orignal_caller })) {
    $self -> { _original_caller } = scalar(caller());
  }

  unless ($#_ == 0) {
    unless (exists($self -> { $key })) {
      $self -> { $key } = preferences -> new_key();
    }

    $self -> { $key } -> append(@_);
  }
  else {
    $self -> { $key } .= shift();

    unless ($self -> { _original_caller } eq "ui") {
      if (exists($self -> { $key . "_callback" })) {
        $self -> { $key . "_callback" } -> ($self -> { $key });
      }
    }

    delete($self -> { _original_caller });
  }

  $self -> { _changed } = 1 if (caller() eq "ui");

  return 1;
}

sub attach_callback {
  my $self = shift();
  my $key = shift();

  unless ($#_ == 0) {
    $self -> { $key } -> attach_callback(@_);
  }
  else {
    $self -> { $key . "_callback" } = shift();
  }

  return 1;
}

sub get {
  my $self = shift();
  my $key = shift();

  return $#_ == -1 ?
    $self -> { $key } :
    $self -> { $key } -> get(@_);
}

sub get_display_elements {
  my ($self) = @_;
  my $file = $self -> { ui } -> { data } . "/display_elements";

  local $/ = undef;

  open(DISPLAY_ELEMENTS, $file) or die("$file: $!\n");

  my @display_elements = split(/\n/, <DISPLAY_ELEMENTS>);

  close(DISPLAY_ELEMENTS);

  return @display_elements;
}

sub index_set {
  my $self = shift();
  my $key = shift();

  unless (exists($self -> { _orignal_caller })) {
    $self -> { _original_caller } = scalar(caller());
  }

  unless ($#_ == 1) {
    unless (exists($self -> { $key })) {
      $self -> { $key } = preferences -> new_key();
    }

    $self -> { $key } -> index_set(@_);
  }
  else {
    my $index = shift();
    my $value = shift();

    $self -> { $key } -> [$index] = $value;

    unless ($self -> { _original_caller } eq "ui") {
      if (exists($self -> { $key . "_callback" })) {
        $self -> { $key . "_callback" } -> ($self -> { $key });
      }
    }

    delete($self -> { _original_caller });
  }

  $self -> { _changed } = 1 if (caller() eq "ui");

  return 1;
}

sub load_defaults {
  my ($self) = @_;

  $self -> set("grouplens_add", []);
  $self -> set("group_display_format", []);
  $self -> set("header_display_format", []);
  $self -> set("include", []);
  $self -> set("interpret", []);
  $self -> set("nnrpaccess", []);
  $self -> set("server", []);
  $self -> set("strip_re_regexp", []);
  $self -> set("strip_was_regexp", []);
  $self -> set("strip_sig_regexp", []);
  $self -> set("unknown", []);

  # need to be changed once there is an interface for the key bindings
  $self -> unset("setkey", "article");
  $self -> unset("setkey", "group");
  $self -> unset("setkey", "readline");

  $self -> unset("compatible_charsets");
  $self -> unset("posting_host");

  $self -> unset("set", "Xbrowser");
  $self -> unset("set", "art_help_line");
  $self -> unset("set", "art_status_line");
  $self -> unset("set", "author_display");
  $self -> unset("set", "charset");
  $self -> unset("set", "custom_headers");
  $self -> unset("set", "custom_sort_order");
  $self -> unset("set", "decode_directory");
  $self -> unset("set", "display_author_realname");
  $self -> unset("set", "display_score");
  $self -> unset("set", "editor_command");
  $self -> unset("set", "failed_posts_file");
  $self -> unset("set", "followup_custom_headers");
  $self -> unset("set", "followup_date_format");
  $self -> unset("set", "generate_email_from");
  $self -> unset("set", "group_dsc_start_column");
  $self -> unset("set", "group_help_line");
  $self -> unset("set", "group_status_line");
  $self -> unset("set", "grouplens_host");
  $self -> unset("set", "grouplens_pseudoname");
  $self -> unset("set", "header_help_line");
  $self -> unset("set", "header_status_line");
  $self -> unset("set", "hostname");
  $self -> unset("set", "inews_program");
  $self -> unset("set", "macro_directory");
  $self -> unset("set", "mail_editor_command");
  $self -> unset("set", "metamail_command");
  $self -> unset("set", "mime_charset");
  $self -> unset("set", "overview_date_format");
  $self -> unset("set", "organization");
  $self -> unset("set", "post_editor_command");
  $self -> unset("set", "post_object");
  $self -> unset("set", "postpone_directory");
  $self -> unset("set", "non_Xbrowser");
  $self -> unset("set", "printer_name");
  $self -> unset("set", "query_reconnect");
  $self -> unset("set", "quote_string");
  $self -> unset("set", "realname");
  $self -> unset("set", "reply_custom_headers");
  $self -> unset("set", "replyto");
  $self -> unset("set", "save_directory");
  $self -> unset("set", "save_posts");
  $self -> unset("set", "save_replies");
  $self -> unset("set", "score_editor_command");
  $self -> unset("set", "scorefile");
  $self -> unset("set", "sendmail_command");
  $self -> unset("set", "server_object");
  $self -> unset("set", "show_descriptions");
  $self -> unset("set", "signature");
  $self -> unset("set", "signoff_string");
  $self -> unset("set", "supersedes_custom_headers");
  $self -> unset("set", "top_status_line");
  $self -> unset("set", "use_blink");
  $self -> unset("set", "use_color");
  $self -> unset("set", "use_uudeview");
  $self -> unset("set", "use_xgtitle");
  $self -> unset("set", "username");

  $self -> index_set("header_display_format", 0, '%F%-5S%G%-5l:[%12r]%t%s');
  $self -> index_set("header_display_format", 1, '%F%G%-5l:[%12r]%t%s');
  $self -> index_set("header_display_format", 2, '%F%-5l:%t%s');
  $self -> index_set("header_display_format", 3, '%F%-5S%-5l:%t%50s %r');

  $self -> index_set("group_display_format", 0, '  %F%-5u  %n%45g%d');
  $self -> index_set("group_display_format", 1, '  %F%-5u  %n%50g%-8l-%h');
  $self -> index_set("group_display_format", 2, '  %F%-5u [%-6t]  %n');

  $self -> set("autobaud", 0);
  $self -> set("ignore_quotes", "^ ?[:>]");
  $self -> set("visible_headers", "From:,Subject:,Newsgroups:,Followup-To:,Reply-To:");

  $self -> set("set", "abort_unmodified_edits", 0);
  $self -> set("set", "auto_mark_article_as_read", 1);
  $self -> set("set", "beep", 1);
  $self -> set("set", "broken_xref", 0);
  $self -> set("set", "cc_followup", 1);
  $self -> set("set", "cc_followup_string", "[This message has also been posted to %n.]");
  $self -> set("set", "cc_post_string", "[This message has also been posted.]");
  $self -> set("set", "check_new_groups", 1);
  $self -> set("set", "color_by_score", 3);
  $self -> set("set", "confirm_actions", 31);
  $self -> set("set", "custom_sort_by_threads", 0);
  $self -> set("set", "display_cursor_bar", 0);
  $self -> set("set", "drop_bogus_groups", 1);
  $self -> set("set", "editor_uses_mime_charset", 0);
  $self -> set("set", "emphasized_text_mask", 1);
  $self -> set("set", "emphasized_text_mode", 3);
  $self -> set("set", "fold_headers", 1);
  $self -> set("set", "followup_string", "In article %m, %r wrote:");
  $self -> set("set", "followup_strip_signature", 1);
  $self -> set("set", "force_authentication", 0);
  $self -> set("set", "generate_date_header", 0);
  $self -> set("set", "generate_message_id", 1);
  $self -> set("set", "grouplens_port", 0);
  $self -> set("set", "hide_pgpsignature", 0);
  $self -> set("set", "hide_quotes", 0);
  $self -> set("set", "hide_signature", 0);
  $self -> set("set", "hide_verbatim_marks", 0);
  $self -> set("set", "highlight_unread_subjects", 1);
  $self -> set("set", "highlight_urls", 1);
  $self -> set("set", "ignore_signature", 0);
  $self -> set("set", "kill_score", -9999);
  $self -> set("set", "lines_per_update", 50);
  $self -> set("set", "mail_editor_is_mua", 0);
  $self -> set("set", "max_low_score", 0);
  $self -> set("set", "max_queued_groups", 20);
  $self -> set("set", "min_high_score", 1);
  $self -> set("set", "mouse", 0);
  $self -> set("set", "netiquette_warnings", 1);
  $self -> set("set", "new_subject_breaks_threads", 0);
  $self -> set("set", "no_autosave", 0);
  $self -> set("set", "no_backups", 0);
  $self -> set("set", "prefer_head", 0);
  $self -> set("set", "process_verbatim_marks", 1);
  $self -> set("set", "query_next_article", 1);
  $self -> set("set", "query_next_group", 1);
  $self -> set("set", "query_read_group_cutoff", 100);
  $self -> set("set", "read_active", 0);
  $self -> set("set", "reject_long_lines", 2);
  $self -> set("set", "reply_string", "In %n, you wrote:");
  $self -> set("set", "scroll_by_page", 0);
  $self -> set("set", "show_article", 0);
  $self -> set("set", "show_thread_subject", 0);
  $self -> set("set", "simulate_graphic_chars", 0);
  $self -> set("set", "smart_quote", 1);
  $self -> set("set", "sorting_method", 3);
  $self -> set("set", "spoiler_char", '*');
  $self -> set("set", "spoiler_display_mode", 1);
  $self -> set("set", "spool_active_file", "data/active");
  $self -> set("set", "spool_activetimes_file", "data/active.times");
  $self -> set("set", "spool_check_up_on_nov", 0);
  $self -> set("set", "spool_inn_root", "/var/lib/news");
  $self -> set("set", "spool_newsgroups_file", "data/newsgroups");
  $self -> set("set", "spool_nov_file", ".overview");
  $self -> set("set", "spool_nov_root", "/var/spool/news");
  $self -> set("set", "spool_root", "/var/spool/news");
  $self -> set("set", "uncollapse_threads", 0);
  $self -> set("set", "unsubscribe_new_groups", 0);
  $self -> set("set", "use_flow_control", 0);
  $self -> set("set", "use_grouplens", 0);
  $self -> set("set", "use_header_numbers", 1);
  $self -> set("set", "use_localtime", 1);
  $self -> set("set", "use_metamail", 1);
  $self -> set("set", "use_mime", 5);
  $self -> set("set", "use_slrnpull", 0);
  $self -> set("set", "use_tilde", 1);
  $self -> set("set", "use_tmpdir", 0);
  $self -> set("set", "warn_followup_to", 1);
  $self -> set("set", "wrap_flags", 3);
  $self -> set("set", "wrap_method", 2);
  $self -> set("set", "write_newsrc_flags", 0);

  foreach ($self -> get_display_elements()) {
    $self -> set("color", $_ . "_fg", "default");
    $self -> set("color", $_ . "_bg", "default");
  }

  foreach ($self -> get_display_elements()) {
    $self -> set("mono", $_ . "_one", "none");
    $self -> set("mono", $_ . "_two", "none");
    $self -> set("mono", $_ . "_three", "none");
    $self -> set("mono", $_ . "_four", "none");
  }

  return 1;
}

sub read {
  my ($self, $rc_file) = @_;

  $self -> load_defaults();
  $Parse::RecDescent::rc_grammar::self = $self;

  #############################################################################

  my $parser = rc_grammar -> new() or return
    $self -> { ui } -> error("rc_grammar: Bad grammar.");

  #############################################################################

  $rc_file = defined($rc_file) ? $rc_file : $self -> { rc_file };

  open(RC_FILE, $rc_file) or return
    $self -> { ui } -> error("Couldn't open '$rc_file' for reading.");

  {
    local $/;

    defined($parser -> rc_file(<RC_FILE>)) or return
      $self -> { ui } -> error("$rc_file: Bad RC file.");
  }

  close(RC_FILE);

  #############################################################################

  $self -> { ui } -> error("Successfully read '$rc_file'.");

  return 1;
}

sub set {
  my $self = shift();
  my $key = shift();

  unless (exists($self -> { _orignal_caller })) {
    $self -> { _original_caller } = scalar(caller());
  }

  unless ($#_ == 0) {
    unless (exists($self -> { $key })) {
      $self -> { $key } = preferences -> new_key();
    }

    $self -> { $key } -> set(@_);
  }
  else {
    $self -> { $key } = shift();

    unless ($self -> { _original_caller } eq "ui") {
      if (exists($self -> { $key . "_callback" })) {
        $self -> { $key . "_callback" } -> ($self -> { $key });
      }
    }

    delete($self -> { _original_caller });
  }

  $self -> { _changed } = 1 if (caller() eq "ui");

  return 1;
}

sub is_callback {
  my ($self, $string) = @_;
  return index($string, "_callback") + 1;
}

sub write {
  my ($self, $rc_file) = @_;

  $rc_file = defined($rc_file) ? $rc_file : $self -> { rc_file };

  open(RC_FILE, ">$rc_file") or return
    $self -> { ui } -> error("Couldn't open '$rc_file' for writing.");

  #############################################################################

  if ($self -> get("autobaud") == 1) {
    print RC_FILE qq(autobaud\n);
  }

  #############################################################################

  my %color = %{$self -> get("color") || {}};
  my %color_seen;

  foreach (sort { $a cmp $b } (keys(%color))) {
    next if ($self -> is_callback($_));

    my $base_name = substr($_, 0, -3);
    my $value_fg = $color{ $base_name . "_fg" } || "default";
    my $value_bg = $color{ $base_name . "_bg" } || "default";

    unless ($color_seen{ $base_name }) {
      print RC_FILE qq(color $base_name "$value_fg" "$value_bg"\n);
    }

    $color_seen{ $base_name }++;
  }

  #############################################################################

  my $compatible_charsets = $self -> get("compatible_charsets");

  if (defined($compatible_charsets)) {
    print RC_FILE qq(compatible_charsets "$compatible_charsets"\n);
  }

  #############################################################################

  foreach (@{$self -> get("grouplens_add") || []}) {
    print RC_FILE qq(grouplens_add "$_"\n);
  }

  #############################################################################

  my @group_display_format = @{$self -> get("group_display_format") || []};

  foreach (0 .. $#group_display_format) {
    print RC_FILE qq(group_display_format $_ "$group_display_format[$_]"\n);
  }

  #############################################################################

  my @header_display_format = @{$self -> get("header_display_format") || []};

  foreach (0 .. $#header_display_format) {
    print RC_FILE qq(header_display_format $_ "$header_display_format[$_]"\n);
  }

  #############################################################################

  my $ignore_quotes = $self -> get("ignore_quotes");

  if (defined($ignore_quotes)) {
    print RC_FILE qq(ignore_quotes "$ignore_quotes"\n);
  }

  #############################################################################

  foreach (@{$self -> get("include") || []}) {
    print RC_FILE qq(include "$_"\n);
  }

  #############################################################################

  foreach (@{$self -> get("interpret") || []}) {
    print RC_FILE qq(interpret "$_"\n);
  }

  #############################################################################

  my %mono = %{$self -> get("mono") || {}};
  my %mono_seen;

  foreach (sort { $a cmp $b } (keys(%mono))) {
    next if ($self -> is_callback($_));

    (my $base_name = $_) =~ s/_(?:one|two|three|four)$//;

    unless ($mono_seen{ $base_name }) {
      my $value_one = $mono{ $base_name . "_one" } || "none";
      my $value_two = $mono{ $base_name . "_two" } || "none";
      my $value_three = $mono{ $base_name . "_three" } || "none";
      my $value_four = $mono{ $base_name . "_four" } || "none";

      print RC_FILE qq(mono $base_name "$value_one" "$value_two" "$value_three" "$value_four"\n);
    }

    $mono_seen{ $base_name }++;
  }

  #############################################################################

  foreach (@{$self -> get("nnrpaccess") || []}) {
    my ($server, $user_name, $password) = split(/ => /, $_);
    print RC_FILE qq(nnrpaccess "$server" "$user_name" "$password"\n);
  }

  #############################################################################

  my $posting_host = $self -> get("posting_host");

  if (defined($posting_host)) {
    print RC_FILE qq(posting_host "$posting_host"\n);
  }

  #############################################################################

  foreach (@{$self -> get("server") || []}) {
    my ($server, $newsrc) = split(/ => /, $_);
    print RC_FILE qq(server "$server" "$newsrc"\n);
  }

  #############################################################################

  my %set = %{$self -> get("set") || {}};

  foreach (sort { $a cmp $b } (keys(%set))) {
    next if ($self -> is_callback($_));

    if (defined($set{ $_ })) {
      if ($set{ $_ } =~ m/^[\d\-\+]+$/) {
        print RC_FILE qq(set $_ $set{ $_ }\n);
      }
      elsif ($set{ $_ } =~ m/[\"\*]/) {
        print RC_FILE qq(set $_ '$set{ $_ }'\n);
      }
      else {
        print RC_FILE qq(set $_ "$set{ $_ }"\n);
      }
    }
  }

  #############################################################################

  foreach (qw(article group readline)) {
    my $type = $_;
    my %setkey = %{$self -> get("setkey", $type) || {}};

    foreach my $key (sort { $a cmp $b } (keys(%setkey))) {
      next if ($self -> is_callback($_));

      foreach (split(/<=>/, $setkey{ $key })) {
        print RC_FILE qq(setkey $type $key "$_"\n);
      }
    }
  }

  #############################################################################

  foreach (qw(re sig was)) {
    my $type = $_;
    my @strip_regex = @{$self -> get("strip_" . $type . "_regexp") || []};

    if (defined($strip_regex[0])) {
      print RC_FILE qq(strip_${type}_regexp "$strip_regex[0]");
      print RC_FILE qq( "$strip_regex[1]") if (defined($strip_regex[1]));
      print RC_FILE qq( "$strip_regex[2]") if (defined($strip_regex[2]));
      print RC_FILE qq( "$strip_regex[3]") if (defined($strip_regex[3]));
      print RC_FILE qq( "$strip_regex[4]") if (defined($strip_regex[4]));
      print RC_FILE qq(\n);
    }
  }

  #############################################################################

  my $visible_headers = $self -> get("visible_headers");

  if (defined($visible_headers)) {
    print RC_FILE qq(visible_headers "$visible_headers"\n);
  }

  #############################################################################

  print RC_FILE qq(\n);

  foreach (@{$self -> get("unknown") || []}) {
    chomp();
    print RC_FILE qq($_\n);
  }

  #############################################################################

  close(RC_FILE);

  $self -> { _changed } = 0 if (caller() eq "ui");

  $self -> { ui } -> error("Successfully wrote '$rc_file'.");

  return 1;
}

sub unset {
  my $self = shift();
  my $key = shift();

  unless (exists($self -> { _orignal_caller })) {
    $self -> { _original_caller } = scalar(caller());
  }

  unless ($#_ == -1) {
    unless (exists($self -> { $key })) {
      $self -> { $key } = preferences -> new_key();
    }

    $self -> { $key } -> unset(@_);
  }
  elsif (exists($self -> { $key })) {
    delete($self -> { $key });

    unless ($self -> { _original_caller } eq "ui") {
      if (exists($self -> { $key . "_callback" })) {
        $self -> { $key . "_callback" } -> ($self -> { $key });
      }
    }

    delete($self -> { _original_caller });
  }

  return 1;
}

1;
